//! Main entry point into the CLI.

use std::{fs, path::PathBuf, process, thread::sleep, time::Duration};

use clap::Parser;
use colored::Colorize;
use dmox::{
    config::{Distance, MAX_MODULE_SIZE},
    Config,
};
use snafu::{ensure, ResultExt, Snafu};

#[derive(Parser)]
#[command(version, about)]
#[allow(clippy::struct_excessive_bools)]
struct Cli {
    /// Forward indexes .fastq.gz file.
    #[arg(long)]
    i1: PathBuf,

    /// Reverse indexes .fastq.gz file.
    #[arg(long)]
    i2: PathBuf,

    /// Forward reads .fastq.gz file.
    #[arg(long)]
    r1: PathBuf,

    /// Reverse reads .fastq.gz file.
    #[arg(long)]
    r2: PathBuf,

    /// Reference 'A' barcode modules.
    #[arg(long)]
    ref_a: PathBuf,

    /// Reference 'B' barcode modules.
    #[arg(long)]
    ref_b: PathBuf,

    /// Reference 'C' barcode modules.
    #[arg(long)]
    ref_c: PathBuf,

    /// Reference 'D' barcode modules.
    #[arg(long)]
    ref_d: PathBuf,

    /// Distance metric to use when matching modules against references.
    #[arg(long, default_value_t = Distance::Levenshtein)]
    distance: Distance,

    /// If set, reject modules matching references
    /// with a distance higher than this threshold.
    #[arg(long)]
    max_distance: Option<usize>,

    /// Schema file mapping sample ids to barcode sample modules with the same letter.
    #[arg(long)]
    schema: PathBuf,

    /// Desired output folder for the resulting sample files.
    /// Created if missing.
    #[arg(long)]
    samples: PathBuf,

    /// Raise to output BX codes as part of demultiplexed fastq identifiers.
    #[arg(
        long,
        num_args(0..=1),
        default_value_t = true,
    )]
    bx: bool,

    /// Raise to output RX codes as part of demultiplexed fastq identifiers.
    #[arg(
        long,
        num_args(0..=1),
        default_value_t = false,
    )]
    rx: bool,

    /// Raise to output QX codes as part of demultiplexed fastq identifiers.
    #[arg(
        long,
        num_args(0..=1),
        default_value_t = false,
    )]
    qx: bool,

    /// Raise to output the last part of original fastq identifiers.
    #[arg(
        long,
        num_args(0..=1),
        default_value_t = false,
    )]
    id_tail: std::primitive::bool,

    /// Desired output compression level (defaults to gzip's default).
    #[arg(long, default_value_t = 6)]
    zlevel: u32,

    /// Where to write the file summarizing assigned / unassigned barcodes.
    #[arg(long)]
    barcodes_table: PathBuf,

    /// Total number of reference barcode modules for each letter.
    #[arg(long, default_value_t = 96)]
    n_modules: u64,

    /// Size of one barcode module.
    #[arg(long, default_value_t = 6)]
    module_size: usize,

    /// Number of threads to spawn for zipping + writing output sample files.
    /// Defaults to the number of available cores on the machine.
    #[arg(long)]
    n_writers: Option<usize>,

    /// Size of memory buffers for accumulation before writing to disk.
    #[arg(long, default_value_t = 1024 * 1024)]
    writers_capacity: usize,

    /// Stop reading input files
    /// if this number of fastq blocks are queued in memory,
    /// still waiting for a writer to pick them up.
    /// Increase to improve performance *provided you have enough RAM*.
    #[arg(long, default_value_t = 256 * 256)]
    max_queued_blocks: usize,

    /// Number of fastq blocks to read before blocking when the above limit is hit.
    /// Don't tweak without profiling performances.
    #[arg(long, default_value_t = 256)]
    readers_channel_capacity: usize,
}

// Catch any error bubbling up from the whole procedure,
// for unified pretty-printing and exit.
fn main() {
    match run() {
        Ok(()) => {
            println!("{}", "✓".bold().green());
            process::exit(0);
        }
        Err(e) => {
            // TODO: error messages may be torn because panics don't lock stderr.
            //   https://github.com/rust-lang/rust/issues/124995
            // In the meantime, just sleep until (hopefully) all threads stop panicking.
            sleep(Duration::from_secs(1));
            eprintln!("{} {e}\n{}", "Error:".bold().red(), "🗙".bold().red());
            process::exit(1);
        }
    }
}

fn run() -> Result<(), Error> {
    // Parse CLI args.
    let Cli {
        i1,
        i2,
        r1,
        r2,
        ref_a,
        ref_b,
        ref_c,
        ref_d,
        distance,
        max_distance,
        schema,
        samples,
        bx,
        rx,
        qx,
        id_tail,
        zlevel,
        barcodes_table,
        n_modules,
        module_size,
        n_writers,
        writers_capacity,
        max_queued_blocks,
        readers_channel_capacity,
    } = Cli::parse();

    // Check config consistency.

    ensure!(
        module_size <= MAX_MODULE_SIZE,
        ConfigErr {
            mess: format!(
                "This version has not been compiled \
                 for barcode modules longer than {MAX_MODULE_SIZE}. \
                 Received module size: {module_size}."
            )
        }
    );

    let n_writers = if let Some(n) = n_writers {
        ensure!(
            n > 0,
            ConfigErr {
                mess: "Cannot use 0 writer thread."
            }
        );
        n
    } else {
        std::thread::available_parallelism()
            .with_context(|_| IoErr {
                doing: "querying the available thread number",
            })?
            .get()
    };

    // Prepare output folder.
    fs::create_dir_all(&samples).with_context(|_| OutputFolderErr { path: &samples })?;

    let config = Config {
        module_size,
        n_modules,
        output_compression_level: zlevel,
        inputs: [i1, i2, r1, r2],
        references: [ref_a, ref_b, ref_c, ref_d],
        distance,
        max_distance,
        schema,
        barcodes_table_file: barcodes_table,
        sample: Box::new(move |id, infix| {
            let mut path = PathBuf::from(&samples);
            path.push(format!("{id}.{infix}.fq.gz"));
            path
        }),
        output_bx: bx,
        output_rx: rx,
        output_qx: qx,
        output_id_tail: id_tail,
        readers_channel_capacity,
        n_writers,
        max_queued_blocks,
        writers_capacity,
    };

    dmox::demultiplex(&config)?;

    Ok(())
}

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
enum Error {
    #[snafu(transparent)]
    Dmox { source: dmox::Error },
    #[snafu(display("Could not create samples folder at {}:\n{source}", path.display()))]
    OutputFolder {
        path: PathBuf,
        source: std::io::Error,
    },
    #[snafu(display("Configuration error: {mess}"))]
    Config { mess: String },
    #[snafu(display("IO error when {doing}: {source}"))]
    Io {
        doing: String,
        source: std::io::Error,
    },
}
