//! Various shared utils around I/O.

use std::{
    fmt::{self, Display},
    fs::File,
    io::{self, BufRead, BufReader, BufWriter, Read, Write},
    path::{Path, PathBuf},
};

use colored::Colorize;
use snafu::{ResultExt, Snafu};

pub(crate) fn open_file(path: &Path) -> Result<File, Error> {
    File::open(path).with_context(|_| OpenErr { path })
}

pub(crate) fn create_file(path: &Path) -> Result<File, Error> {
    File::create(path).with_context(|_| CreateErr { path })
}

//--------------------------------------------------------------------------------------------------

/// Associate this value to errors values to contextualize I/O error messages.
#[derive(Debug, Clone)]
pub struct Context {
    pub(crate) path: PathBuf,
    pub(crate) i_line: u64,
}

impl Display for Context {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let Self { path, i_line } = self;
        writeln!(
            f,
            "In file {}:{}:",
            path.display().to_string().blue(),
            i_line.to_string().green()
        )
    }
}

//--------------------------------------------------------------------------------------------------
// Produce context values from readers.

/// A reader aware of filename and line number,
/// useful to produce error contexts.
pub(crate) struct ContextualizedReader<R: Read> {
    reader: BufReader<R>,
    path: PathBuf,
    i_line: u64, // Manually handled by consumer.
}

impl<R: Read> ContextualizedReader<R> {
    pub(crate) fn new(path: PathBuf, reader: BufReader<R>) -> Self {
        Self {
            reader,
            path,
            i_line: 1,
        }
    }

    pub(crate) fn bump_line(&mut self) {
        self.i_line += 1;
    }

    pub(crate) fn i_line(&self) -> u64 {
        self.i_line
    }

    pub(crate) fn read_until(&mut self, byte: u8, buf: &mut Vec<u8>) -> Result<usize, Error> {
        self.reader
            .read_until(byte, buf)
            .with_context(|_| ReadErr { context: self })
    }
}

impl ContextualizedReader<File> {
    pub(crate) fn from_file(path: &Path) -> Result<Self, Error> {
        let file = open_file(path)?;
        let reader = BufReader::new(file);
        Ok(Self::new(path.into(), reader))
    }
}

// The whole point.
impl<R: Read> From<&ContextualizedReader<R>> for Context {
    fn from(r: &ContextualizedReader<R>) -> Self {
        Self {
            path: r.path.clone(),
            i_line: r.i_line,
        }
    }
}

// Convenience to ease snafu selector syntax.
impl<R: Read> From<ContextualizedReader<R>> for Context {
    fn from(r: ContextualizedReader<R>) -> Self {
        (&r).into()
    }
}
impl<R: Read> From<&mut ContextualizedReader<R>> for Context {
    fn from(r: &mut ContextualizedReader<R>) -> Self {
        (&*r).into()
    }
}

//--------------------------------------------------------------------------------------------------
// Same idea for writers.

/// A writer aware of filename and line number,
/// useful to produce error contexts.
pub(crate) struct ContextualizedWriter<W: Write> {
    writer: W,
    path: PathBuf,
    pub(crate) i_line: u64,
}

impl<W: Write> ContextualizedWriter<W> {
    pub(crate) fn new(path: PathBuf, writer: W) -> Self {
        Self {
            writer,
            path,
            i_line: 1,
        }
    }
    pub(crate) fn bump_line(&mut self) {
        self.i_line += 1;
    }
}

// Delegate trait to inner writer.
impl<W: Write> Write for ContextualizedWriter<W> {
    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
        self.writer.write(buf)
    }
    fn flush(&mut self) -> io::Result<()> {
        self.writer.flush()
    }
}

impl ContextualizedWriter<BufWriter<File>> {
    pub(crate) fn from_file(path: &Path) -> Result<Self, Error> {
        let file = create_file(path)?;
        let writer = BufWriter::new(file);
        Ok(Self::new(path.into(), writer))
    }
}

// The whole point.
impl<W: Write> From<&ContextualizedWriter<W>> for Context {
    fn from(w: &ContextualizedWriter<W>) -> Self {
        Self {
            path: w.path.clone(),
            i_line: w.i_line,
        }
    }
}

// Convenience to ease snafu selector syntax.
impl<W: Write> From<ContextualizedWriter<W>> for Context {
    fn from(w: ContextualizedWriter<W>) -> Self {
        (&w).into()
    }
}
impl<W: Write> From<&mut ContextualizedWriter<W>> for Context {
    fn from(w: &mut ContextualizedWriter<W>) -> Self {
        (&*w).into()
    }
}

//--------------------------------------------------------------------------------------------------

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(display("Could not open file {}:\n{source}", path.display()))]
    Open { source: io::Error, path: PathBuf },
    #[snafu(display("Could not read file {}:\n{source}", path.display()))]
    ReadFile { source: io::Error, path: PathBuf },
    #[snafu(display("Could not create file {}:\n{source}", path.display()))]
    Create { source: io::Error, path: PathBuf },
    #[snafu(display("{context}IO error:\n{source}."))]
    Read {
        source: std::io::Error,
        context: Context,
    },
}

// Snafu convenience.
impl From<&Context> for Context {
    fn from(c: &Context) -> Self {
        c.clone()
    }
}
