//! Use this module to parse raw fastq blocks from input files.
//!
//! The provided reader adapter feeds on successive lines,
//! assuming the fastq blocks succeed as 4 rotations of:
//!
//!  "@" <id_head> <id_tail> // <- (split on the first space separator)
//!  <read>
//!  "+" <id>?   // <- Same <id> but optional (actually ignored).
//!  <quality>   // <- Same size as <read>.
//!
//! On each iteration, it consumes 4 more lines
//! to provides the parsed block as an *owned buffer*.
//!
//! The id is split because the BX code needs to be inserted between head & tail.

use std::{io::Read, mem, ops::Range};

use snafu::{ensure, Snafu};

use crate::{
    io::{self, Context, ContextualizedReader},
    short::ful,
};

/// One yielded parsed fastq block.
#[derive(Clone)]
pub(crate) struct Block {
    pub(crate) i_line: u64,     // Block starting line in fastq file.
    pub(crate) buffer: Vec<u8>, // All collected bytes.
    // Index into the various parts.
    pub(crate) id_head: Range<usize>, // (includes leading '@' and space separator)
    pub(crate) id_tail: Range<usize>, // (includes trailing newline) (actually ignored)
    pub(crate) read: Range<usize>,    // (idem)
    pub(crate) separator: Range<usize>, // (idem)
    pub(crate) quality: Range<usize>, // (idem (added if missing from input eof))
}

/// Wrap reader to recycle the single allocated yielded block.
pub(crate) struct BlocksReader<R: Read> {
    reader: ContextualizedReader<R>,
    block: Block,
}

impl Block {
    /// Initialize empty block.
    pub(crate) fn new() -> Self {
        Self {
            i_line: 0,
            buffer: Vec::new(),
            id_head: 0..0,
            id_tail: 0..0,
            read: 0..0,
            separator: 0..0,
            quality: 0..0,
        }
    }

    /// Useful for error context updates.
    pub(crate) fn n_lines() -> u64 {
        4
    }

    /// Paste id head and tail into a single slice.
    pub(crate) fn id(&self) -> &[u8] {
        &self.buffer[self.id_head.start..self.id_tail.end]
    }

    /// Paste read and quality into a single slice.
    pub(crate) fn read_quality(&self) -> &[u8] {
        &self.buffer[self.read.start..self.quality.end]
    }

    /// Forget about this block, but don't release the underlying allocation.
    fn clear(&mut self) {
        let Self {
            i_line,
            buffer,
            id_head,
            id_tail,
            read,
            separator,
            quality,
        } = self;
        *i_line = 0;
        buffer.clear();
        *id_head = 0..0;
        *id_tail = 0..0;
        *read = 0..0;
        *separator = 0..0;
        *quality = 0..0;
    }
}

impl<R: Read> BlocksReader<R> {
    pub(crate) fn new(reader: ContextualizedReader<R>) -> Self {
        Self {
            reader,
            block: Block::new(),
        }
    }

    /// Assuming the block has been cleared, parse a new id into it.
    /// False on EOF.
    fn parse_id_line(&mut self) -> Result<bool, ParseError> {
        let Self { reader, block: b } = self;
        reader.bump_line();

        // First read down to the first separator.
        let n = reader.read_until(b' ', &mut b.buffer)?;
        if n == 0 {
            return Ok(false); // EOF reached if nothing was read.
        }
        b.id_head = 0..n;

        ensure!(b.buffer.starts_with(b"@"), IdErr { id: b.id_head() });

        // Then read down to EOL.
        let n = reader.read_until(b'\n', &mut b.buffer)?;
        let s = b.id_head.end;
        b.id_tail = s..(s + n);

        ensure!(
            b.buffer.ends_with(b"\n"),
            IncompleteErr {
                missing: "identifier",
            }
        );

        Ok(true)
    }

    /// Read and parse next block so it's ready for reading after that.
    /// False on EOF.
    pub(crate) fn parse_next_block(&mut self) -> Result<bool, Error> {
        let mut parse = || -> Result<bool, ParseError> {
            // Start with the id line.
            self.block.clear();
            self.block.i_line = self.reader.i_line();
            if !self.parse_id_line()? {
                return Ok(false); // Reached EOF.
            }
            let Self { block: b, reader } = self;

            // If any, then the 'read', 'separator' ('+' line) and 'quality' must be present.
            for (line, missing) in [
                (&mut b.read, "read"),
                (&mut b.separator, "'+' separator"),
                (&mut b.quality, "quality"),
            ] {
                let s = b.buffer.len();
                let n = reader.read_until(b'\n', &mut b.buffer)?;
                ensure!(n > 0, IncompleteErr { missing });
                *line = s..(s + n);
            }

            // Rough consistency check: ignore possible id repetition.
            ensure!(b.separator()[0] == b'+', SepErr { sep: b.separator() });

            // Insert quality trailing newline if missing at EOF.
            #[allow(clippy::range_plus_one)]
            if b.buffer.last().unwrap() != &b'\n' {
                b.buffer.push(b'\n');
                b.quality = (b.quality.start)..(b.quality.end + 1);
            }
            ensure!(
                b.read.len() == b.quality.len(),
                LengthsErr {
                    read: b.read(),
                    quality: b.quality(),
                }
            );

            Ok(true)
        };

        // Upgrade error with context.
        parse().map_err(|source| Error::Parse {
            context: (&self.reader).into(),
            source,
        })
    }

    /// Retrieve the block just parsed.
    pub(crate) fn block(&self) -> &Block {
        &self.block
    }

    /// Useful to recycle blocks allocations.
    pub(crate) fn swap(&mut self, block: &mut Block) {
        mem::swap(&mut self.block, block);
    }
}

//--------------------------------------------------------------------------------------------------
// Generate getters for the various block parts.

macro_rules! get {

    // Get raw slice.
    ($($part:ident)+) => {$(
        pub(crate) fn $part(&self) -> &[u8] {
            &self.buffer[self.$part.clone()]
        }
    )+};

    // Trim trailing char.
    (trimmed: $($part:ident)+) => {$(
        pub(crate) fn $part(&self) -> &[u8] {
            let r = &self.$part;
            let (s, e) = (r.start, r.end);
            &self.buffer[s..(e - 1)]
        }
    )+};

}

impl Block {
    get! { separator }
    get! { trimmed: id_head id_tail read quality }
}

//==================================================================================================
#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(transparent)]
    Io { source: io::Error },
    #[snafu(display("{context}{source}"))]
    Parse {
        context: Context,
        source: ParseError,
    },
}

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum ParseError {
    #[snafu(transparent)]
    Io { source: io::Error },
    #[snafu(display("File ended before {missing:?} line could be read."))]
    Incomplete { missing: &'static str },
    #[snafu(display("Invalid fastq identifier: {:?}.", ful(id)))]
    Id { id: Vec<u8> },
    #[snafu(display("Invalid fastq '+' separator: {:?}.", ful(sep)))]
    Sep { sep: Vec<u8> },
    #[snafu(display("The quality size is {} for a read of size {}:\
                     \n     read: {}\
                     \n  quality: {}", read.len(), quality.len(), ful(read), ful(quality)))]
    Lengths { read: Vec<u8>, quality: Vec<u8> },
}

impl<R: Read> From<&BlocksReader<R>> for io::Context {
    fn from(br: &BlocksReader<R>) -> Self {
        (&br.reader).into()
    }
}

impl<R: Read> From<&mut BlocksReader<R>> for io::Context {
    fn from(br: &mut BlocksReader<R>) -> Self {
        (&*br).into()
    }
}
